from pathlib import Path

from langchain.agents import Tool
from llama_index import (
    LLMPredictor,
    PromptHelper,
    ServiceContext,
    SimpleDirectoryReader,
    StorageContext,
    VectorStoreIndex,
)
from llama_index.indices.composability import ComposableGraph
from llama_index.indices.keyword_table import SimpleKeywordTableIndex
from llama_index.indices.query.query_transform.base import DecomposeQueryTransform
from llama_index.query_engine.transform_query_engine import TransformQueryEngine
from llama_index.vector_stores import ChromaVectorStore

from servifai.memory import ChromaDB


class VectorKnowledgeBase:
    """Vector Index Knowledge Base"""

    def __init__(self, vector_indices, index_summaries, service_context):
        self._indices = vector_indices
        self._summaries = index_summaries
        self._service_context = service_context

    def as_tool(self, title: str) -> Tool:
        """access vector index as a tool

        Args:
            title (str): index title

        Returns:
            Tool: index as tool
        """
        index = self._indices[title]
        summary = self._summaries[title]
        query_engine = index.as_query_engine(
            service_context=self._service_context, similarity_top_k=3
        )
        return Tool(
            name=f"Knowledge Vector Index {title}",
            func=lambda q: str(query_engine.query(q)),
            description=f"useful for answering queries only regarding {summary}",
            return_direct=True,
        )


class KnowledgeGraphs:
    """Graph Knowledge Base"""

    def __init__(self, vector_indices, index_summaries, service_context, llm_predictor):
        self._indices = vector_indices
        self._summaries = index_summaries
        self._service_context = service_context
        self._llm_pred = llm_predictor
        self._graph = None
        self._custom_query_engines = {}
        self._query_engine = self._create_graph_qe()

    def _create_graph_qe(self):
        self._graph = ComposableGraph.from_indices(
            SimpleKeywordTableIndex,
            [index for _, index in self._indices.items()],
            [summary for _, summary in self._summaries.items()],
            max_keywords_per_chunk=50,
        )
        decompose_transform = DecomposeQueryTransform(self._llm_pred, verbose=True)

        for index in self._indices.values():
            query_engine = index.as_query_engine(service_context=self._service_context)
            transform_extra_info = {"index_summary": index.index_struct.summary}
            tranformed_query_engine = TransformQueryEngine(
                query_engine,
                decompose_transform,
                transform_metadata=transform_extra_info,
            )
            self._custom_query_engines[index.index_id] = tranformed_query_engine

        self._custom_query_engines[
            self._graph.root_index.index_id
        ] = self._graph.root_index.as_query_engine(
            retriever_mode="simple",
            response_mode="tree_summarize",
            service_context=self._service_context,
        )
        return self._graph.as_query_engine(
            custom_query_engines=self._custom_query_engines
        )

    def as_tool(self, about: str) -> Tool:
        """access knowledge graph as tool

        Args:
            about (str): about the data

        Returns:
            Tool: graph as tool
        """
        return Tool(
            name="Knowledge Graph Index",
            func=lambda q: str(self._query_engine.query(q)),
            description=f"useful for answering queries that require comparing/contrasting or analyzing over multiple sources about {about}",
            return_direct=True,
        )


class KnowledgeBase:
    """Combined Knowledge Base including vector and graph indices"""

    def __init__(self, vdb_dir, data_dir, about, memoryconfigs, oaillm):
        self.vdb_dir = vdb_dir
        self.data_dir = data_dir
        self.about = about
        self.max_input_size = int(memoryconfigs["max_input_size"])
        self.num_outputs = int(memoryconfigs["num_outputs"])
        self.max_chunk_overlap = float(memoryconfigs["max_chunk_overlap"])
        self.chunk_size_limit = int(memoryconfigs["chunk_size_limit"])
        self.llm_predictor = LLMPredictor(llm=oaillm.model)
        self.vectordb = ChromaDB(self.vdb_dir)
        self.vectorstore = self._initiate_vectorstore()
        self.service_context = self._initiate_contexts()[0]
        self.storage_context = self._initiate_contexts()[1]
        self.titles = []
        self.indices = self._create_indices()[0]
        self.index_summaries = self._create_indices()[1]
        self.qe_tools = self._get_query_engine_tools()

    def _initiate_vectorstore(self):
        return ChromaVectorStore(chroma_collection=self.vectordb.collection)

    def _initiate_contexts(self):
        prompt_helper = PromptHelper(
            self.max_input_size,
            self.num_outputs,
            self.max_chunk_overlap,
            chunk_size_limit=self.chunk_size_limit,
        )
        self.service_context = ServiceContext.from_defaults(
            llm_predictor=self.llm_predictor,
            prompt_helper=prompt_helper,
            embed_model=self.vectordb.embed,
        )
        self.storage_context = StorageContext.from_defaults(
            vector_store=self.vectorstore
        )
        return self.service_context, self.storage_context

    def _create_indices(self):
        """
        Loads the data, chunks it, create embedding for each chunk
        and then stores the embedding to a vector database.
        Currently it supports pdf and docx file formats - two
        most used formats as personal unstructured document reader.

        Args:
            data_dir (str): the directory containing the data
        """
        exts = [".pdf", ".docx"]
        docs = {}
        indices = {}
        index_summaries = {}

        for ext in exts:
            for dbb in Path(self.data_dir).glob(f"*{ext}"):
                title = " ".join(dbb.stem.split("-"))
                self.titles.append(title)
                docs[title] = SimpleDirectoryReader(
                    input_files=[str(dbb)],
                    recursive=True,
                    exclude_hidden=True,
                    required_exts=[ext],
                ).load_data()
                indices[title] = VectorStoreIndex.from_documents(
                    docs[title],
                    service_context=self.service_context,
                    storage_context=self.storage_context,
                )
                index_summaries[
                    title
                ] = f"individual source generally about {self.about} and particularly on {title}"
        return indices, index_summaries

    def _get_query_engine_tools(self):
        qe_tools = []
        for title in self.titles:
            vectorkb_tool = VectorKnowledgeBase(
                self.indices, self.index_summaries, self.service_context
            ).as_tool(title)
            qe_tools.append(vectorkb_tool)

        kg_tool = KnowledgeGraphs(
            self.indices, self.index_summaries, self.service_context, self.llm_predictor
        ).as_tool(self.about)
        qe_tools.append(kg_tool)
        return qe_tools

    def as_tool(self):
        """access the combined knowledge base as a tool

        Returns:
            list: list of knowledge indices as tools
        """
        return self.qe_tools
